# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SUSPECT utility functions."""

from functools import wraps
import warnings
from suspect.pyomo.util import (
    model_variables,
    model_constraints,
    model_objectives,
    bounds_and_expr
)


def deprecated(new_func):
    """Mark a function as deprecated."""
    def _deprecated_decorator(func):
        @wraps(func)
        def _wrapper(*args, **kwargs):
            warnings.warn(
                '{} has been renamed to {}.'.format(
                    func.__name__,
                    new_func,
                ),
                DeprecationWarning,
            )
            return func(*args, **kwargs)
        return _wrapper
    return _deprecated_decorator
